"""Formula 8.13 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM2
from blueprints.validations import raise_if_negative


class Form8Dot13AdditionalShearReinforcement(Formula):
    """Class representing formula 8.13 for the calculation of the minimum additional shear reinforcement in the anchorage zones where transverse
    compression is not present for straight anchorage lengths, in the direction perpendicular to the tension face.
    """

    label = "8.13"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        a_s: MM2,
        n_2: DIMENSIONLESS,
    ) -> None:
        """[:math:`A_{sh}`] Minimum additional shear reinforcement in the anchorage zones where transverse compression is not present for straight
        anchorage lengths, in the direction perpendicular to the tension face [:math:`mm²`].

        NEN-EN 1992-1-1+C2:2011 art.8.8(6) - Formula (8.12)

        Parameters
        ----------
        a_s: MM2
            [:math:`A_{s}`] Cross sectional area of reinforcement [:math:`mm²`].
        n_2: DIMENSIONLESS
            [:math:`n_{2}`] Number of bars anchored in each layer [-].
        """
        super().__init__()
        self.a_s = a_s
        self.n_2 = n_2

    @staticmethod
    def _evaluate(
        a_s: MM2,
        n_2: DIMENSIONLESS,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a_s=a_s, n_2=n_2)
        return 0.25 * a_s * n_2

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 8.13."""
        return LatexFormula(
            return_symbol=r"A_{sv}",
            result=f"{self:.2f}",
            equation=r"0.25 \cdot A_s \cdot n_2",
            numeric_equation=rf"0.25 \cdot {self.a_s:.2f} \cdot {self.n_2:.2f}",
            comparison_operator_label="=",
        )



"""Formula 5.14 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot14SlendernessRatio(Formula):
    """Class representing formula 5.14 for the calculation of the slenderness ratio, :math:`λ`."""

    label = "5.14"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        l_0: M,
        i: M,
    ) -> None:
        """[:math:`λ`] Slenderness ratio [:math:`-`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.3.2(1) - Formula (5.14)

        Parameters
        ----------
        l_0 : M
            [:math:`l_{0}`] Effective length [:math:`m`].
            Use your own implementation of this value or use :class: `Form5Dot15EffectiveLengthBraced`
            or :class: `Form5Dot15EffectiveLengthUnbraced`.
        i : M
            [:math:`i`] Radius of gyration of the uncracked concrete section [:math:`m`].
        """
        super().__init__()
        self.l_0 = l_0
        self.i = i

    @staticmethod
    def _evaluate(
        l_0: M,
        i: M,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(
            l_0=l_0,
            i=i,
        )
        return l_0 / i

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.14."""
        return LatexFormula(
            return_symbol=r"λ",
            result=f"{self:.3f}",
            equation=r"\frac{l_0}{i}",
            numeric_equation=rf"\frac{{{self.l_0:.3f}}}{{{self.i:.3f}}}",
            comparison_operator_label="=",
        )



"""Module for unit conversions inside of Blueprints."""

import math

# <editor-fold desc="Forces conversion">
KN_TO_N = 1e3
N_TO_KN = 1e-3
MM3_TO_M3 = 1e-9
MPA_TO_KPA = 1e3
GPA_TO_MPA = 1e3
# </editor-fold>

# <editor-fold desc="Section modulus conversion">
CM3_TO_MM3 = 1e3
# </editor-fold>

# <editor-fold desc="Moments conversion">
NMM_TO_KNM = 1e-6
# </editor-fold>

# <editor-fold desc="Area conversion">
MM2_TO_M2 = 1e-6
# </editor-fold>

# <editor-fold desc="Rotation conversion">
RAD_TO_MRAD = 1e3
MRAD_TO_RAD = 1e-3
RAD_TO_DEG = 180 / math.pi
DEG_TO_RAD = math.pi / 180
# </editor-fold>


# <editor-fold desc="Length conversion">
M_TO_MM = 1e3
MM_TO_M = 1e-3



"""Formula 3.24 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA
from blueprints.validations import raise_if_negative


class Form3Dot24And25IncreasedCharacteristicCompressiveStrength(Formula):
    """Class representing formula 3.24 and 3.25 for the calculation of the increased characteristic compressive strength due to enclosed concrete."""

    label = "3.24 - 3.25"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ck: MPA,
        sigma_2: MPA,
    ) -> None:
        """[:math:`f_{ck,c}`] Increased characteristic compressive strength due to enclosed concrete [:math:`MPa`].

        NEN-EN 1992-1-1+C2:2011 art.3.1.9(2) - Formula (3.24 and 3.25)

        Parameters
        ----------
        f_ck : MPA
            [:math:`f_{ck}`] Characteristic compressive strength concrete [:math:`MPa`].
            Valid range: :math:`f_{ck} ≤ 90 MPa`.
        sigma_2 : MPA
            [:math:`σ_2`] Effective compressive stress in transverse direction [:math:`MPa`].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ck = f_ck
        self.sigma_2 = sigma_2

    @staticmethod
    def _evaluate(
        f_ck: MPA,
        sigma_2: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(f_ck=f_ck, sigma_2=sigma_2)
        if sigma_2 <= 0.05 * f_ck:
            return f_ck * (1.000 + 5.0 * sigma_2 / f_ck)
        return f_ck * (1.125 + 2.5 * sigma_2 / f_ck)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.24 and 3.25."""
        if self.sigma_2 <= 0.05 * self.f_ck:
            return LatexFormula(
                return_symbol=r"f_{ck,c}",
                result=f"{self:.2f}",
                equation=r"f_{ck} \cdot (1.000 + 5.0 \cdot \sigma_2 / f_{ck})",
                numeric_equation=rf"{self.f_ck:.2f} \cdot (1.000 + 5.0 \cdot {self.sigma_2:.2f} / {self.f_ck:.2f})",
                comparison_operator_label="=",
            )
        return LatexFormula(
            return_symbol=r"f_{ck,c}",
            result=f"{self:.2f}",
            equation=r"f_{ck} \cdot (1.125 + 2.5 \cdot \sigma_2 / f_{ck})",
            numeric_equation=rf"{self.f_ck:.2f} \cdot (1.125 + 2.5 \cdot {self.sigma_2:.2f} / {self.f_ck:.2f})",
            comparison_operator_label="=",
        )



"""Module containing definitions of soil types and soil materials."""

from dataclasses import dataclass
from enum import Enum


class SoilMainTypes(Enum):
    """Soil types based on grain size, representing the main portion of the soil mix."""

    GRAVEL = "Gravel"
    SAND = "Sand"
    SILT = "Silt"
    LOAM = "Loam"
    CLAY = "Clay"
    PEAT = "Peat"


class SoilMinorCompositions(Enum):
    """Soil minor compositions, representing the minor portion of the soil mix."""

    CLEAN = "Clean"

    SLITGHTLY_SILTY = "Slightly silty"
    VERY_SILTY = "Very silty"

    SLIGHTLY_CLAYEY = "Slightly clayey"
    VERY_CLAYEY = "Very clayey"

    SLIGHTLY_SANDY = "Slightly sandy"
    VERY_SANDY = "Very sandy"

    ORGANIC = "Organic"


class SoilConsolidationStates(Enum):
    """Consolidation states of the soil."""

    UNCONSOLIDATED = "Unconsolidated"
    NORMALLY_CONSOLIDATED = "Normally consolidated"
    SLIGHTLY_OVERCONSOLIDATED = "Slightly consolidated"
    MODERATELY_OVERCONSOLIDATED = "Moderately overconsolidated"
    HIGHLY_OVERCONSOLIDATED = "Highly consolidated"


class SoilConsistencies(Enum):
    """Consistencies of the soil."""

    # Clay-like consistencies
    VERY_SOFT = "Very soft"
    SOFT = "Soft"
    MEDIUM_STIFF = "Medium stiff"
    STIFF = "Stiff"
    HARD = "Hard"

    # Sand-like consistencies
    VERY_LOOSE = "Very loose"
    LOOSE = "Loose"
    MEDIUM_DENSE = "Medium dense"
    DENSE = "Dense"
    VERY_DENSE = "Very dense"


@dataclass
class SoilType:
    """Class representing a soil type.

    Parameters
    ----------
    main_type : SoilMainTypes
        The soil type of the main portion of the soil mix.
    minor_composition : SoilMinorCompositions, optional
        The composition of the minor portion of the soil mix, by default None.
    consolidation_state : SoilConsolidationStates, optional
        The consolidation state of the soil, by default None.
    consistency : SoilConsistencies, optional
        The consistency of the soil, by default None.
    description : str, optional
        The custom description of the soil, by default None.
    """

    main_type: SoilMainTypes
    minor_composition: SoilMinorCompositions | None = None
    consolidation_state: SoilConsolidationStates | None = None
    consistency: SoilConsistencies | None = None
    description: str | None = None

    @property
    def default_description(self) -> str:
        """The default description of the soil type based on the `main_type`, `minor_composition`, `consolidation_state`, and `consistency`."""
        default_description = self.main_type.value

        for prop in [self.minor_composition, self.consolidation_state, self.consistency]:
            if prop is not None:
                assert isinstance(prop, Enum)
                default_description += ", " + prop.value
        return default_description



"""Formula 5.17 from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

import math

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, KN_M2, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot17EffectiveLengthBucklingLoad(Formula):
    """Class representing formula 5.17 for the calculation of effective length of unbraced members, in the
    case where criteria (2) and (3) do not apply such as by variable loading, :math:`l_0`.
    """

    label = "5.17"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, ei: KN_M2, n_b: KN) -> None:
        """[:math:`l_{0}`] Effective length for unbraced members [:math:`m`].

        NEN-EN 1992-1-1+C2:2011 art.5.8.3.2(6) - Formula (5.17)

        Parameters
        ----------
        ei : KN_M2
            [:math:`EI`] is a representative bending stiffness [:math:`kN/m^2`].
        n_b : KN
            [:math:`N_{b}`] is the buckling load expressed in terms of EI (in equation (5.14) i should correspond
            to this EI). [:math:`kN`].
        """
        super().__init__()
        self.ei = ei
        self.n_b = n_b

    @staticmethod
    def _evaluate(ei: KN_M2, n_b: KN) -> M:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(ei=ei, n_b=n_b)
        raise_if_less_or_equal_to_zero(n_b=n_b)
        return math.pi * math.sqrt(ei / n_b)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.17."""
        return LatexFormula(
            return_symbol=r"l_0",
            result=f"{self:.3f}",
            equation=r"\pi \cdot \sqrt{\frac{EI}{N_{b}}}",
            numeric_equation=rf"\pi \cdot \sqrt{{\frac{{{self.ei}}}{{{self.n_b}}}}}",
            comparison_operator_label="=",
        )



"""Formula 3.21 and 3.22 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA


class Form3Dot21And22EffectiveStrength(Formula):
    """[:math:`η`] Class representing formula 3.21 and 3.22 for the calculation of the factor for the effective strength [:math:`-`]."""

    label = "3.21 - 3.22"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ck: MPA,
    ) -> None:
        r"""[:math:`η`] Factor effective strength [:math:`-`].

        NEN-EN 1992-1-1+C2:2011 art.3.1.7(3) - Formula (3.21) and (3.22)

        Parameters
        ----------
        f_ck : MPA
            [:math:`f_{ck}`] Characteristic compressive strength concrete [:math:`MPa`].
            Valid range: :math:`f_{ck} \leq 90 MPa`.

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        f_ck: MPA,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_ck <= 50:
            return 1.0
        if f_ck <= 90:
            return 1.0 - (f_ck - 50) / 200
        raise ValueError(f"Invalid f_ck: {f_ck}. Maximum of f_ck is 90 MPa")

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.21 and 3.22."""
        return LatexFormula(
            return_symbol=r"\eta",
            result=f"{self:.3f}",
            equation=r"1.0" if self.f_ck <= 50 else r"1.0 - (f_{ck} - 50) / 200",
            numeric_equation=r"1.0" if self.f_ck <= 50 else rf"1.0 - ({self.f_ck:.3f} - 50) / 200",
            comparison_operator_label="=",
        )



"""Formula 8.4 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_5 import (
    Form8Dot5ProductAlphas235,
)
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets
from blueprints.type_alias import MM, RATIO
from blueprints.validations import raise_if_negative


class Form8Dot4DesignAnchorageLength(Formula):
    """Class representing formula 8.4 for the calculation of the design anchorage length :math:`l_{bd}` [mm]."""

    label = "8.4"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        alpha_1: RATIO,
        alpha_2: RATIO,
        alpha_3: RATIO,
        alpha_4: RATIO,
        alpha_5: RATIO,
        l_b_rqd: MM,
        l_b_min: MM,
        min_product_alpha_2_3_5: RATIO | None = None,
    ) -> None:
        """[:math:`l_{bd}`] Design anchorage length [:math:`mm`].

        NEN-EN 1992-1-1+C2:2011 art.8.4.4(1) - Formula (8.4)

        Parameters
        ----------
        alpha_1 : RATIO
            [:math:`α_{1}`] Coefficient for the effect of the form of the bars assuming adequate cover (see figure 8.1) [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1.0` for straight bars in tension.

            :math:`= 1.0 if c_{d} <= 3 ⋅ Ø` for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).

            :math:`= 0.7 if c_{d} > 3 ⋅ Ø` for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).

            Note: see figure 8.3 for values of :math:`c_{d}`.
        alpha_2 : RATIO
            [:math:`α_{2}`] Coefficient for the effect of minimum concrete cover (see figure 8.3) [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1 - 0.15 ⋅ (c_{d} - Ø) / Ø <= 1` with a minimum of 0.7 for straight bars in tension.

            :math:`= 1 - 0.15 ⋅ (c_{d} - 3 ⋅ Ø) / Ø <= 1` with a minimum of 0.7 for bars other than
            straight in tension (see figure 8.1 (b), (c) and (d)).

            Note: see figure 8.3 for values of :math:`c_{d}`.
        alpha_3 : RATIO
            [:math:`α_{3}`] Coefficient for the effect of confinement by transverse reinforcement [-].

            :math:`= 1.0` for bars in compression.

            :math:`= 1 - K ⋅ λ <= 1` with a minimum of 0.7 for bars in tension.

            Where: :math:`λ = (ΣA_{st} - ΣA_{st,min}) / A_{s}`.

            Where: :math:`ΣA_{st,min}` = cross-sectional area of the minimum transverse
            reinforcement :math:`= 0,25 ⋅ A_{s}` for beams and 0 for slabs.

            Note: see figure 8.4 for values of :math:`K, A_{s} and A_{st}`.
        alpha_4 : RATIO
            [:math:`α_{4}`] Coefficient for the influence of one or more welded transverse bars :math:`(Ø_{t} > 0,6 Ø)` along the design anchorage
            length :math:`l_{bd}` (see 8.6) [-].

            :math:`= 0.7` for all types, position and size as specified in figure 8.6 (e) in both tension and compression.

        alpha_5 : RATIO
            [:math:`α_{5}`] Coefficient for the effect of the pressure transverse to the plane of splitting
            along the design anchorage length :math:`l_{bd}` (see 8.6) [-].

            :math:`= 1 - 0.04 ⋅ p <= 1` with a minimum of 0.7 for all types of anchorage in compression.

            Where: p = transverse pressure at ultimate limit state along :math:`l_{bd}` [MPa].
        l_b_rqd: MM
            [:math:`l_{b,rqd}`] Basic required anchorage length, for anchoring the force As⋅σsd in a straight bar assuming constant
            bond stress (formula 8.3) [:math:`mm`].

            Use your own implementation for this value or use the :class:`Form8Dot3RequiredAnchorageLength` class.
        l_b_min : MM
            [:math:`l_{b,min}`] Minimum anchorage length if no other limitation is applied [:math:`mm`].

            :math:`= max(0.3 ⋅ l_{b,rqd}, 10 ⋅ Ø, 100)` for tension anchorage (formula 8.6).
            :math:`= max(0.6 ⋅ l_{b,rqd}, 10 ⋅ Ø, 100)` for compression anchorage (formula 8.7).

            Use your own implementation of this formula or use the :class:`Form8Dot6MinimumTensionAnchorage` class for tension or
            :class:`Form8Dot7MinimumCompressionAnchorage` for compression.
        min_product_alpha_2_3_5: RATIO | None
            Minimum value of the product of factors alpha 2, alpha 3 and alpha 5.
            When this argument is None, :class: `Form8Dot5ProductAlphas235` is used for this condition.
            When this argument is given, the condition :math: max((`α_{2}``α_{3}``α_{5}`) >= min_product_alpha_2_3_5 is used.

        Notes
        -----
        NEN-EN 1992-1-1+C2:2011 art.8.4.4(1) - Formula (8.5) prescribes that :math:`(α_{2} ⋅ α_{3} ⋅ α_{5}) >= 0.7`.
        """
        super().__init__()
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.alpha_3 = alpha_3
        self.alpha_4 = alpha_4
        self.alpha_5 = alpha_5
        self.l_b_rqd = l_b_rqd
        self.l_b_min = l_b_min
        self.min_product_alpha_2_3_5 = min_product_alpha_2_3_5

    @staticmethod
    def _evaluate(
        alpha_1: RATIO,
        alpha_2: RATIO,
        alpha_3: RATIO,
        alpha_4: RATIO,
        alpha_5: RATIO,
        l_b_rqd: MM,
        l_b_min: MM,
        min_product_alpha_2_3_5: RATIO | None = None,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        )

        if min_product_alpha_2_3_5 is None:
            product_alphas_2_3_5: float = Form8Dot5ProductAlphas235(alpha_2=alpha_2, alpha_3=alpha_3, alpha_5=alpha_5)
        else:
            product_alphas_2_3_5 = max(alpha_2 * alpha_3 * alpha_5, min_product_alpha_2_3_5)

        return max(alpha_1 * alpha_4 * product_alphas_2_3_5 * l_b_rqd, l_b_min)

    def latex(self) -> LatexFormula:
        """Returns a LatexFormula representation of the formula."""
        return LatexFormula(
            return_symbol=r"l_{bd}",
            result=f"{self:.2f}",
            equation=latex_max_curly_brackets(r"\alpha_1 \cdot \alpha_2 \cdot \alpha_3 \cdot \alpha_4 \cdot \alpha_5 \cdot l_{b,rqd}", r"l_{b,min}"),
            numeric_equation=latex_max_curly_brackets(
                rf"{self.alpha_1} \cdot {self.alpha_2} \cdot {self.alpha_3} \cdot {self.alpha_4} \cdot {self.alpha_5} \cdot {self.l_b_rqd:.2f}",
                self.l_b_min,
            ),
            comparison_operator_label="=",
        )



"""Formula 5.10b from NEN-EN 1992-1-1+C2:2011: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot10bRedistributionOfMomentsUpperFck:
    r"""Class representing formula 5.10b for the redistribution of moments in continuous beams or slabs when :math:`f_{ck} > 50 MPa`."""

    label = "5.10b"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(self, delta: DIMENSIONLESS, k3: DIMENSIONLESS, k4: DIMENSIONLESS, xu: M, d: M) -> None:
        r"""[:math:`δ`] Redistribution of moments in continuous beams or slabs when :math:`f_{ck} > 50 MPa`.

        NEN-EN 1992-1-1+C2:2011 art.5.5(4) - Formula (5.10b)

        Parameters
        ----------
        delta : DIMENSIONLESS
            [:math:`δ`] is the ratio of the redistributed moment to the elastic moment.
        k3 : DIMENSIONLESS
            [:math:`k3`] is a coefficient for redistribution.
        k4 : DIMENSIONLESS
            [:math:`k4`] is a coefficient for redistribution.
        xu : M
            [:math:`x_u`] is the depth of the compression zone in the ultimate limit state after redistribution.
        d : M
            [:math:`d`] is the effective depth of the section.
        """
        super().__init__()
        self.delta = delta
        self.k3 = k3
        self.k4 = k4
        self.xu = xu
        self.d = d
        raise_if_less_or_equal_to_zero(delta=delta, k3=k3, k4=k4, xu=xu, d=d)

    @property
    def left_hand_side(self) -> DIMENSIONLESS:
        """Calculate the left hand side of the comparison.

        Returns
        -------
        DIMENSIONLESS
            Left hand side of the comparison.
        """
        return self.delta

    @property
    def right_hand_side(self) -> DIMENSIONLESS:
        """Calculate the right hand side of the comparison.

        Returns
        -------
        DIMENSIONLESS
            Right hand side of the comparison.
        """
        raise_if_less_or_equal_to_zero(d=self.d)
        raise_if_less_or_equal_to_zero(xu=self.xu)
        return self.k3 + self.k4 * (self.xu / self.d)

    @property
    def ratio(self) -> DIMENSIONLESS:
        """Ratio between left hand side and right hand side of the comparison, commonly referred to as unity check."""
        return self.left_hand_side / self.right_hand_side

    def __bool__(self) -> bool:
        """Evaluates the comparison, for more information see the __init__ method."""
        return self.left_hand_side >= self.right_hand_side

    def __str__(self) -> str:
        """Return the result of the comparison."""
        return self.latex().complete

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.10b."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"\delta \geq k_3 + k_4 \frac{x_u}{d}",
            numeric_equation=rf"{self.left_hand_side:.3f} \geq {self.k3} + {self.k4} \frac{{{self.xu}}}{{{self.d}}}",
            comparison_operator_label=r"\rightarrow",
        )



"""Definitions for the nominal concrete cover check according to art. 4.4.1 (Concrete cover) from NEN-EN 1992-1-1+C2:2011."""

from enum import Enum


class AbrasionClass(Enum):
    """Enum representing the abrasion class of the concrete surface.

    According to art. 4.4.1.2 (13) from NEN-EN 1992-1-1+C2:2011
    """

    NA = "Not applicable"
    XM1 = "XM1"
    XM2 = "XM2"
    XM3 = "XM3"


class CastingSurface(Enum):
    """Enum representing the casting surface of the concrete.

    According to art. 4.4.1.3 (4) from NEN-EN 1992-1-1+C2:2011
    """

    PERMANENTLY_EXPOSED = "Permanently exposed"
    FORMWORK = "Formwork"
    PREPARED_GROUND = "Prepared ground (including blinding)"
    DIRECTLY_AGAINST_SOIL = "Directly against soil"



"""Formula 3.13 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DAYS


class Form3Dot13CoefficientTimeAutogeneShrinkage(Formula):
    """Class representing formula 3.13, which calculates the coefficient dependent on time for the autogene shrinkage."""

    source_document = NEN_EN_1992_1_1_C2_2011
    label = "3.13"

    def __init__(
        self,
        t: DAYS,
    ) -> None:
        """[βas(t)] Coefficient dependent on time in days for autogene shrinkage [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(6) - Formula (3.13)

        Parameters
        ----------
        t : DAYS
            [t] Time in days [days].

        Returns
        -------
        None
        """
        super().__init__()
        self.t = t

    @staticmethod
    def _evaluate(
        t: DAYS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if t < 0:
            raise ValueError(f"Invalid t: {t}. t cannot be negative")
        return 1 - np.exp(-0.2 * t**0.5)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.13."""
        return LatexFormula(
            return_symbol=r"\beta_{as}(t)",
            result=f"{self:.3f}",
            equation=r"1 - \exp(-0.2 \cdot t^{0.5})",
            numeric_equation=rf"1 - \exp(-0.2 \cdot {self.t:.2f}^{{0.5}})",
            comparison_operator_label="=",
        )



"""Formula 3.7 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.validations import raise_if_negative


class Form3Dot7NonLinearCreepCoefficient(Formula):
    """Class representing formula 3.7 for the calculation of the non-linear creep coefficient."""

    label = "3.7"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        phi_inf_t0: float,
        k_sigma: float,
    ) -> None:
        """[φnl(∞,t0)] The non-linear creep coefficient [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(4) - Formula (3.7)

        Parameters
        ----------
        phi_inf_t0 : float
            [φ(∞, t0)] Creep coefficient if high accuracy is not required use figure 3.1 and/or use appendix B [-].
        k_sigma : float
            [kσ] Stress-strength ratio (σc / fck(t0)) [-].

        Returns
        -------
        None
        """
        super().__init__()
        self.phi_inf_t0 = phi_inf_t0
        self.k_sigma = k_sigma

    @staticmethod
    def _evaluate(
        phi_inf_t0: float,
        k_sigma: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(phi_inf_t0=phi_inf_t0, k_sigma=k_sigma)
        return phi_inf_t0 * np.exp(1.5 * (k_sigma - 0.45))

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.7."""
        return LatexFormula(
            return_symbol=r"\varphi_{nl}(\infty, t_0)",
            result=f"{self:.3f}",
            equation=r"\varphi(\infty, t_0) \cdot \exp( 1.5 ( k_{\sigma} - 0.45))",
            numeric_equation=rf"{self.phi_inf_t0:.3f} \cdot \exp( 1.5 ( {self.k_sigma:.3f} - 0.45))",
            comparison_operator_label="=",
        )



"""Formula 6.71 from NEN-EN 1992-1-1+C2:2011: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot71CriteriaBasedOnStressRangeLHS(Formula):
    """Class representing Left Hand Side of formula 6.71 for the calculation of the fatigue criteria based on stress range."""

    label = "6.71 (LHS)"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        gamma_f_fat: DIMENSIONLESS,
        delta_sigma_s_equ_n_star: MPA,
    ) -> None:
        """[:math:`Δσ_{Ed}`] Loading side of equation [:math:`MPa`].

        NEN-EN 1993-1-1+C2:2011 art.6.8.5 - Formula (6.71)

        Parameters
        ----------
        gamma_f_fat : DIMENSIONLESS
            [:math:`γ_{F,fat}`] Partial factor for fatigue actions [:math:`-`].
        delta_sigma_s_equ_n_star : MPA
            [:math:`Δσ_{s,equ}(N*)`] Damage equivalent stress range for types of reinforcement and considering number of cycles N* [:math:`MPa`].

        Returns
        -------
        None
        """
        super().__init__()
        self.gamma_f_fat = gamma_f_fat
        self.delta_sigma_s_equ_n_star = delta_sigma_s_equ_n_star

    @staticmethod
    def _evaluate(
        gamma_f_fat: DIMENSIONLESS,
        delta_sigma_s_equ_n_star: MPA,
    ) -> MPA:
        """Evaluates the left hand side formula, for more information see the __init__ method."""
        raise_if_negative(gamma_f_fat=gamma_f_fat, delta_sigma_s_equ_n_star=delta_sigma_s_equ_n_star)
        return gamma_f_fat * delta_sigma_s_equ_n_star

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for left hand side formula 6.71."""
        return LatexFormula(
            return_symbol=r"\Delta \sigma_{Ed}",
            result=f"{self:.3f}",
            equation=r"\gamma_{F,fat} \cdot \Delta \sigma_{s,equ} (N^*)",
            numeric_equation=rf"{self.gamma_f_fat:.3f} \cdot {self.delta_sigma_s_equ_n_star:.3f}",
            comparison_operator_label=r"=",
        )


class Form6Dot71CriteriaBasedOnStressRangeRHS(Formula):
    """Class representing Right Hand Side of formula 6.71 for the calculation of the fatigue criteria based on stress range."""

    label = "6.71 (RHS)"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        delta_sigma_rsk_n_star: MPA,
        gamma_s_fat: DIMENSIONLESS,
    ) -> None:
        """[:math:`Δσ_{Rd}`] Resistance side of equation [:math:`MPa`].

        NEN-EN 1993-1-1+C2:2011 art.6.8.5 - Formula (6.71)

        Parameters
        ----------
        delta_sigma_rsk_n_star : MPA
            [:math:`Δσ_{Rsk}(N*)`] Stress range at N* cycles from the S-N curve in Figure 6.30 [:math:`MPa`].
        gamma_s_fat : DIMENSIONLESS
            [:math:`γ_{S,fat}`] Partial factor for reinforcing or prestressing steel under fatigue loading [:math:`-`].

        Returns
        -------
        None
        """
        super().__init__()
        self.delta_sigma_rsk_n_star = delta_sigma_rsk_n_star
        self.gamma_s_fat = gamma_s_fat

    @staticmethod
    def _evaluate(gamma_s_fat: DIMENSIONLESS, delta_sigma_rsk_n_star: MPA) -> MPA:
        """Evaluates the right hand side formula, for more information see the __init__ method."""
        raise_if_negative(delta_sigma_rsk_n_star=delta_sigma_rsk_n_star)
        raise_if_less_or_equal_to_zero(gamma_s_fat=gamma_s_fat)
        return delta_sigma_rsk_n_star / gamma_s_fat

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for right hand side formula 6.71."""
        return LatexFormula(
            return_symbol=r"\Delta \sigma_{Rd}",
            result=f"{self:.3f}",
            equation=r"\frac{\Delta \sigma_{Rsk} (N^*)}{\gamma_{s,fat}}",
            numeric_equation=rf"\frac{{{self.delta_sigma_rsk_n_star:.3f}}}{{{self.gamma_s_fat:.3f}}}",
            comparison_operator_label=r"=",
        )


class Form6Dot71CriteriaBasedOnStressRange:
    """Class representing formula 6.71 for the calculation of the fatigue criteria based on stress range."""

    label = "6.71"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        gamma_f_fat: DIMENSIONLESS,
        delta_sigma_s_equ_n_star: MPA,
        delta_sigma_rsk_n_star: MPA,
        gamma_s_fat: DIMENSIONLESS,
    ) -> None:
        """[:math:`CHECK`] Criteria met, based on damage accumulation.

        NEN-EN 1993-1-1+C2:2011 art.6.8.5 - Formula (6.71)

        Parameters
        ----------
        gamma_f_fat : DIMENSIONLESS
            [:math:`γ_{F,fat}`] Partial factor for fatigue actions [:math:`-`].
        delta_sigma_s_equ_n_star : MPA
            [:math:`Δσ_{s,equ}(N*)`] Damage equivalent stress range for types of reinforcement and considering number of cycles N* [:math:`MPa`].
        delta_sigma_rsk_n_star : MPA
            [:math:`Δσ_{Rsk}(N*)`] Stress range at N* cycles from the S-N curve in Figure 6.30 [:math:`MPa`].
        gamma_s_fat : DIMENSIONLESS
            [:math:`γ_{S,fat}`] Partial factor for reinforcing or prestressing steel under fatigue loading [:math:`-`].

        Returns
        -------
        None
        """
        self.gamma_f_fat = gamma_f_fat
        self.delta_sigma_s_equ_n_star = delta_sigma_s_equ_n_star
        self.delta_sigma_rsk_n_star = delta_sigma_rsk_n_star
        self.gamma_s_fat = gamma_s_fat

    @property
    def left_hand_side(self) -> MPA:
        """Calculate the left hand side of the equation.

        Returns
        -------
            MPA: Left hand side, loading side of the equation
        """
        return Form6Dot71CriteriaBasedOnStressRangeLHS(
            gamma_f_fat=self.gamma_f_fat,
            delta_sigma_s_equ_n_star=self.delta_sigma_s_equ_n_star,
        )

    @property
    def right_hand_side(self) -> MPA:
        """Calculate the left hand side of the equation.

        Returns
        -------
            MPA: Right hand side, resistance side of the equation
        """
        return Form6Dot71CriteriaBasedOnStressRangeRHS(
            gamma_s_fat=self.gamma_s_fat,
            delta_sigma_rsk_n_star=self.delta_sigma_rsk_n_star,
        )

    @property
    def ratio(self) -> DIMENSIONLESS:
        """Ratio between left hand side and right hand side of the formula, commonly referred to as unity check."""
        return self.left_hand_side / self.right_hand_side

    def __bool__(self) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        lhs = Form6Dot71CriteriaBasedOnStressRangeLHS(gamma_f_fat=self.gamma_f_fat, delta_sigma_s_equ_n_star=self.delta_sigma_s_equ_n_star)
        rhs = Form6Dot71CriteriaBasedOnStressRangeRHS(gamma_s_fat=self.gamma_s_fat, delta_sigma_rsk_n_star=self.delta_sigma_rsk_n_star)
        return lhs <= rhs

    def __str__(self) -> str:
        """Return the result of the formula."""
        return self.latex().complete

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 6.71."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"\gamma_{F,fat} \cdot \Delta \sigma_{s,equ} (N^*) \leq \frac{\Delta \sigma_{Rsk} (N^*)}{\gamma_{s,fat}}",
            numeric_equation=(
                rf"{self.gamma_f_fat:.3f} \cdot {self.delta_sigma_s_equ_n_star:.3f} "
                rf"\leq \frac{{{self.delta_sigma_rsk_n_star:.3f}}}{{{self.gamma_s_fat:.3f}}}"
            ),
            comparison_operator_label=r"\rightarrow",
        )



"""Formula 6.5 from NEN-EN 1993-1-1+C2+A1:2016: Chapter 6 - Ultimate limit state."""

# pylint: disable=arguments-differ
from blueprints.codes.eurocode.nen_en_1993_1_1_c2_a1_2016 import NEN_EN_1993_1_1_C2_A1_2016
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, RATIO
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot5UnityCheckTensileStrength(Formula):
    """Class representing formula 6.5 for the unity check for tensile strength."""

    label = "6.5"
    source_document = NEN_EN_1993_1_1_C2_A1_2016

    def __init__(
        self,
        n_ed: KN,
        n_t_rd: KN,
    ) -> None:
        """[N_ed/N_t_rd] Unity check for tensile strength of an element in tension.

        NEN-EN 1993-1-1+C2+A1:2016 art.6.2.3(1) - Formula (6.5)

        Parameters
        ----------
        n_ed : KN
            [NEd] Design value of the normal tensile force [kN].
        n_t_rd : KN
            [Nt,Rd] Design value of the resistance against tenslie force [kN].
        """
        super().__init__()
        self.n_ed = n_ed
        self.n_t_rd = n_t_rd

    @staticmethod
    def _evaluate(
        n_ed: KN,
        n_t_rd: KN,
    ) -> RATIO:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(n_t_rd=n_t_rd)
        raise_if_negative(n_ed=n_ed)
        return n_ed / n_t_rd

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 6.5."""
        return LatexFormula(
            return_symbol=r"N_{Ed}/N_{t,Rd}",
            result=f"{self:.2f}",
            equation=r"N_{Ed} / N_{t,Rd}",
            numeric_equation=rf"{self.n_ed:.2f} / {self.n_t_rd:.2f}",
            comparison_operator_label="=",
        )



"""Geometry operations module."""

import math
from enum import Enum, auto

from shapely import Point

from blueprints.type_alias import RAD


class CoordinateSystemOptions(Enum):
    """Enum of the coordinate system options."""

    XY = auto()
    XZ = auto()
    YZ = auto()


def calculate_rotation_angle(
    start_point: Point,
    end_point: Point,
    coordinate_system: CoordinateSystemOptions = CoordinateSystemOptions.XY,
) -> RAD:
    """
    Calculates rotation of an end point in relation to the start point in a given plane/coordinate system [rad].

    - Start point lies in the origin center of the quadrant.
    - The rotation is calculated in the given plane in relation to the start point.
    - The rotation is calculated in the range [0, 2*pi].
    - The rotation is calculated in the counter-clockwise direction.

    Parameters
    ----------
    start_point: Point
        Starting point that will be used as reference.
    end_point: Point
        End point
    coordinate_system: CoordinateSystemOptions
        Desired plane that will be used as reference to calculate the rotation. Standard is XY-plane.

    Returns
    -------
    RAD
        rotation of end point relative to the start point in radians in the given plane [rad].

    Raises
    ------
    ValueError
        If start_point and end_point are the same.
        If start_point or end_point do not have z value when rotation angle in XZ or YZ plane is requested.
    """
    if list(start_point.coords) == list(end_point.coords):
        msg = f"Start and end point can't be equal. {start_point=} | {end_point=}"
        raise ValueError(msg)

    if coordinate_system != CoordinateSystemOptions.XY and (not start_point.has_z or not end_point.has_z):
        msg = f"Coordinate system '{coordinate_system}' requires z value in both points."
        raise ValueError(msg)

    # Map coordinate system to the corresponding attributes
    coordinates_map = {
        CoordinateSystemOptions.XY: ("x", "y"),
        CoordinateSystemOptions.XZ: ("x", "z"),
        CoordinateSystemOptions.YZ: ("y", "z"),
    }
    horizontal_attr, vertical_attr = coordinates_map[coordinate_system]

    # Extract the coordinates for the specified plane
    horizontal_1, vertical_1 = getattr(start_point, horizontal_attr), getattr(start_point, vertical_attr)
    horizontal_2, vertical_2 = getattr(end_point, horizontal_attr), getattr(end_point, vertical_attr)

    # Calculate the differences in coordinates
    dx = horizontal_2 - horizontal_1
    dy = vertical_2 - vertical_1

    # Calculate the angle using atan2 for correct quadrant determination
    angle = math.atan2(dy, dx)

    # Normalize angle to be in the range [0, 2*pi]
    if angle < 0:
        angle += 2 * math.pi

    return angle



"""Formula 8.3 from NEN-EN 1992-1-1+C2:2011: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_fraction
from blueprints.type_alias import MM, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot3RequiredAnchorageLength(Formula):
    """Class representing formula 8.3 for the calculation of the basic required anchorage length, assuming constant bond stress :math:`f_{bd}`."""

    label = "8.3"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        diameter: MM,
        sigma_sd: MPA,
        f_bd: MPA,
    ) -> None:
        r"""[:math:`l_{b,rqd}`] Basic required anchorage length, for anchoring the force :math:`A_{s} \\cdot σ_{sd}` in a straight bar assuming
        constant bond stress :math:`f_{bd}`. [mm].

        NEN-EN 1992-1-1+C2:2011 art.8.4.3(2) - Formula (8.3)

        Parameters
        ----------
        diameter: MM
            [:math:`Ø`] Diameter of the bar [mm].
        sigma_sd: MPA
            [:math:`σ_{sd}`] design stress of the bar at the position from where the anchorage is measured from [MPa].
        f_bd: MPA
            [:math:`f_{bd}`] Design value ultimate bond stress [MPa].
            Use your own implementation for this value or use the Form8Dot2UltimateBondStress class.
        """
        super().__init__()
        self.diameter = diameter
        self.sigma_sd = sigma_sd
        self.f_bd = f_bd

    @staticmethod
    def _evaluate(
        diameter: MM,
        sigma_sd: MPA,
        f_bd: MPA,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(diameter=diameter, sigma_sd=sigma_sd)
        raise_if_less_or_equal_to_zero(f_bd=f_bd)
        return (diameter / 4) * (sigma_sd / f_bd)

    def latex(self) -> LatexFormula:
        """Returns a LatexFormula object for this formula."""
        latex_diameter = r"Ø"
        latex_sigma_sd = r"\sigma_{sd}"
        latex_f_bd = r"f_{bd}"
        return LatexFormula(
            return_symbol=r"l_{b,rqd}",
            result=f"{self:.2f}",
            equation=rf"{latex_fraction(latex_diameter, 4)} \cdot {latex_fraction(latex_sigma_sd, latex_f_bd)}",
            numeric_equation=rf"{latex_fraction(self.diameter, 4)} \cdot {latex_fraction(self.sigma_sd, self.f_bd)}",
            comparison_operator_label="=",
        )



"""Cross-section shapes for reinforced concrete sections."""

import math
from dataclasses import dataclass
from typing import Protocol

from shapely import Point, Polygon

from blueprints.type_alias import MM, MM2


@dataclass(frozen=True)
class CircularCrossSection:
    """
    Class to represent a circular cross-section using shapely for geometric calculations.

    Parameters
    ----------
    diameter : MM
        The diameter of the circular cross-section [mm].
    x : MM
        The x-coordinate of the circle's center.
    y : MM
        The y-coordinate of the circle's center.
    """

    diameter: MM
    x: MM
    y: MM

    def __post_init__(self) -> None:
        """Post-initialization to validate the diameter."""
        if self.diameter <= 0:
            msg = f"Diameter must be a positive value, but got {self.diameter}"
            raise ValueError(msg)

    @property
    def radius(self) -> MM:
        """
        Calculate the radius of the circular cross-section [mm].

        Returns
        -------
        MM
            The radius of the circle.
        """
        return self.diameter / 2.0

    @property
    def geometry(self) -> Polygon:
        """
        Shapely Polygon representing the circular cross-section.

        Returns
        -------
        Polygon
            The shapely Polygon representing the circle.
        """
        return self.centroid.buffer(self.radius)

    @property
    def area(self) -> MM2:
        """
        Calculate the area of the circular cross-section [mm²].

        Returns
        -------
        MM2
            The area of the circle.
        """
        return math.pi * self.radius**2.0

    @property
    def perimeter(self) -> MM:
        """
        Calculate the perimeter (circumference) of the circular cross-section [mm].

        Returns
        -------
        MM
            The perimeter of the circle.
        """
        return math.pi * self.radius * 2.0

    @property
    def centroid(self) -> Point:
        """
        Get the centroid of the circular cross-section.

        Returns
        -------
        Point
            The centroid of the circle.
        """
        return Point(self.x, self.y)

    @property
    def vertices(self) -> list[Point]:
        """
        Vertices of the circular cross-section.

        Returns
        -------
        list[Point]
            The vertices of the circle.
        """
        return [Point(x, y) for x, y in self.geometry.exterior.coords]


@dataclass(frozen=True)
class RectangularCrossSection:
    """
    Class to represent a rectangular cross-section for geometric calculations.

    Parameters
    ----------
    width : MM
        The width of the rectangular cross-section.
    height : MM
        The height of the rectangular cross-section.
    x : MM
        The x-coordinate of the centroid of the rectangle. Default is 0.
    y : MM
        The y-coordinate of the centroid of the rectangle. Default is 0.
    """

    width: MM
    height: MM
    x: MM = 0
    y: MM = 0

    @property
    def geometry(self) -> Polygon:
        """
        Shapely Polygon representing the rectangular cross-section. Defines the coordinates of the rectangle based on width, height, x,
        and y. Counter-clockwise order.

        Returns
        -------
        Polygon
            The shapely Polygon representing the rectangle.
        """
        left_lower = (self.x - self.width / 2, self.y - self.height / 2)
        right_lower = (self.x + self.width / 2, self.y - self.height / 2)
        right_upper = (self.x + self.width / 2, self.y + self.height / 2)
        left_upper = (self.x - self.width / 2, self.y + self.height / 2)
        return Polygon(
            [
                left_lower,
                right_lower,
                right_upper,
                left_upper,
            ]
        )

    @property
    def area(self) -> MM2:
        """
        Calculate the area of the rectangular cross-section.

        Returns
        -------
        MM2
            The area of the rectangle.
        """
        return self.geometry.area

    @property
    def perimeter(self) -> MM:
        """
        Calculate the perimeter of the rectangular cross-section.

        Returns
        -------
        MM
            The perimeter of the rectangle.
        """
        return self.geometry.length

    @property
    def centroid(self) -> Point:
        """
        Get the centroid of the rectangular cross-section.

        Returns
        -------
        Point
            The centroid of the rectangle.
        """
        return self.geometry.centroid

    @property
    def vertices(self) -> list[Point]:
        """
        Vertices of the rectangular cross-section. Counter-clockwise order starting from the bottom-left corner.

        Returns
        -------
        list[Point]
            The vertices of the rectangle.
        """
        return [Point(x, y) for x, y in self.geometry.exterior.coords]


class CrossSection(Protocol):
    """Protocol for a cross-section."""

    @property
    def geometry(self) -> Polygon:
        """Shapely Polygon representing the cross-section."""

    @property
    def area(self) -> MM2:
        """Area of the cross-section [mm²]."""

    @property
    def perimeter(self) -> MM:
        """Perimeter of the cross-section [mm]."""

    @property
    def centroid(self) -> Point:
        """Centroid of the cross-section [mm]."""

    @property
    def vertices(self) -> list[Point]:
        """Vertices of the cross-section."""



"""Formula 9.4 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailing of members and particular rules."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DEG, DIMENSIONLESS, MM, MM2
from blueprints.validations import raise_if_greater_than_90, raise_if_negative


class Form9Dot4ShearReinforcementRatio(Formula):
    """Class representing the formula 9.4 for the calculation of the shear reinforcement ratio."""

    label = "9.4"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        a_sw: MM2,
        s: MM,
        b_w: MM,
        alpha: DEG,
    ) -> None:
        """[ρw] Shear reinforcement ratio [-].

        NEN-EN 1992-1-1+C2:2011 art.9.2.2(5) - Formula (9.4)

        Parameters
        ----------
        a_sw: MM2
            [Asw] Area of shear reinforcement within length s [mm²].
        s: MM
            [s] The spacing between shear reinforcement along the longitudinal axis of the element [mm].
        b_w: MM
            [bw] The width of the web of the element [mm].
        alpha: DEG
            [α] The angle between the shear reinforcement and the longitudinal axis of the beam (see 9.2.2(1)) [deg].
        """
        super().__init__()
        self.a_sw = a_sw
        self.s = s
        self.b_w = b_w
        self.alpha = alpha

    @staticmethod
    def _evaluate(
        a_sw: MM2,
        s: MM,
        b_w: MM,
        alpha: DEG,
    ) -> DIMENSIONLESS:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(a_sw=a_sw, s=s, b_w=b_w, alpha=alpha)
        raise_if_greater_than_90(alpha=alpha)

        # Convert the angle from degrees to radians
        alpha_radians = np.deg2rad(alpha)

        return a_sw / (s * b_w * np.sin(alpha_radians))

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 9.4."""
        return LatexFormula(
            return_symbol=r"\rho_w",
            result=f"{self:.6f}",
            equation=r"\frac{A_{sw}}{s \cdot b_w \cdot sin(\alpha)}",
            numeric_equation=rf"\frac{{{self.a_sw:.2f}}}{{{self.s:.2f} \cdot {self.b_w:.2f} \cdot sin({self.alpha:.2f})}}",
            comparison_operator_label="=",
        )



"""Formula 9.1N from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailing of members and particular rules."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_fraction, latex_max_curly_brackets
from blueprints.type_alias import MM, MM2, MPA
from blueprints.validations import raise_if_negative


class Form9Dot1nMinimumTensileReinforcementBeam(Formula):
    """Class representing the formula 9.1N for the calculation of minimum tensile reinforcement area in longitudinal direction for beams."""

    label = "9.1N"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        f_ctm: MPA,
        f_yk: MPA,
        b_t: MM,
        d: MM,
    ) -> None:
        """[As,min] Calculates the minimum required tensile reinforcement area in longitudinal direction for beams [mm²].

        NEN-EN 1992-1-1+C2:2011 art.9.2.1.1(1) - Formula (9.1N)

        Notes
        -----
        As,min is no less than 0,0013 * bt * d

        Parameters
        ----------
        f_ctm: MPA
            [fctm] Mean axial tensile stress concrete [MPa].
            Should be determined with respect to the relevant strength class according to Table 3.1
        f_yk: MPA
            [fyk] Characteristic yield strength reinforcement steel [MPa].
        b_t: MM
            [bt] Mean width of the concrete tension zone, for T-beams with a flange under compression only the width of the web is considered for
            calculating bt [mm].
        d: MM
            [d] Effective height of the cross-section [mm].
        """
        super().__init__()
        self.f_ctm = f_ctm
        self.f_yk = f_yk
        self.b_t = b_t
        self.d = d

    @staticmethod
    def _evaluate(
        f_ctm: MPA,
        f_yk: MPA,
        b_t: MM,
        d: MM,
    ) -> MM2:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(f_ctm=f_ctm, f_yk=f_yk, b_t=b_t, d=d)
        return max(0.26 * (f_ctm / f_yk) * b_t * d, 0.0013 * b_t * d)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 9.1N."""
        fraction = latex_fraction(numerator=f"{self.f_ctm:.2f}", denominator=f"{self.f_yk:.2f}")
        return LatexFormula(
            return_symbol=r"A_{s,min}",
            result=f"{self:.2f}",
            equation=latex_max_curly_brackets(
                rf"0.26 \cdot {latex_fraction(numerator=r'f_{ctm}', denominator=r'f_{yk}')} \cdot b_t \cdot d",
                r"0.0013 \cdot b_t \cdot d",
            ),
            numeric_equation=latex_max_curly_brackets(
                rf"0.26 \cdot {fraction} \cdot {self.b_t:.2f} \cdot {self.d:.2f}",
                rf"0.0013 \cdot {self.b_t:.2f} \cdot {self.d:.2f}",
            ),
            comparison_operator_label="=",
        )



"""Module for the concrete exposure classes
according to Table 4.1 from NEN-EN 1992-1-1: Chapter 4 - Durability and cover to reinforcement.
"""

import re
from abc import abstractmethod
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from typing import Self, TypeVar

from blueprints.utils.abc_enum_meta import ABCEnumMeta

T = TypeVar("T", bound="Exposure")


@total_ordering
class Exposure(Enum, metaclass=ABCEnumMeta):
    """Parent class for individual exposure classes.

    This class handles the ordering/comparison operations, that's why it is decorated with total_ordering (As recommended by PEP8).
    On top of that, it handles a couple of methods which will be used by its subclasses.
    """

    def __eq__(self, other: object) -> bool:
        """Definition of '==' operator for the comparison of the severity of the exposure classifications.

        Parameters
        ----------
        self : Exposure/ subclass of Exposure
            First argument for the comparison.
        other : object
            Second argument for the comparison.

        Raises
        ------
        TypeError
            If different types are being compared.

        Returns
        -------
        Boolean
            True if both arguments are of the same severity (In this case they will both be literally the same).
        """
        if not isinstance(other, self.__class__):
            raise TypeError("Only the same exposure class types can be compared with each other!")
        _self_severity = int(self.value[-1]) if self.value != "Not applicable" else 0
        _other_severity = int(other.value[-1]) if other.value != "Not applicable" else 0
        return _self_severity == _other_severity

    def __gt__(self, other: Self) -> bool:
        """Definition of '>' operator for the comparison of the severity of the exposure classifications.

        Parameters
        ----------
        self : Exposure/ subclass of Exposure
            First argument for the comparison.
        other : object
            Second argument for the comparison.

        Raises
        ------
        TypeError
            If different types are being compared.

        Returns
        -------
        Boolean
            True if the first argument is more severe than the second argument.
        """
        if isinstance(other, self.__class__):
            _self_severity = int(self.value[-1]) if self.value != "Not applicable" else 0
            _other_severity = int(other.value[-1]) if other.value != "Not applicable" else 0
            return _self_severity > _other_severity
        raise TypeError("Only the same exposure class types can be compared with each other!")

    @classmethod
    def options(cls: type[T]) -> list[str]:
        """Return all the possible options within a subclass.

        Returns
        -------
        list[str]
            all the possible class designations within a specific exposure class
        """
        return [m._value_ for m in cls.__members__.values()]

    @staticmethod
    @abstractmethod
    def exposure_class_description() -> str:
        """Description of subclasses to be implemented in each subclass.

        Returns
        -------
        str
            description of the specific exposure class
        """

    @abstractmethod
    def description_of_the_environment(self) -> str:
        """Description of the environment based on the instance.

        Returns
        -------
        str
            description of the environment based on the instance
        """

    @classmethod
    def snake_case(cls) -> str:
        """Converts the name of a subclass to snake_case which can be used in a parametrization class.

        Returns
        -------
        str
            the name of a subclass in snake_case
        """
        return re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__name__).lower()


@total_ordering
class CarbonationBase(Exposure):
    """Enum Class which indicates the classification of corrosion induced by carbonation."""


@total_ordering
class ChlorideBase(Exposure):
    """Enum Class which indicates the classification of corrosion induced by chlorides other than by sea water."""


@total_ordering
class ChlorideSeawaterBase(Exposure):
    """Enum Class which indicates the classification of corrosion induced by chlorides from sea water."""


@total_ordering
class FreezeThawBase(Exposure):
    """Enum Class which indicates the classification of freeze/thaw attack with or without de-icing agents."""


@total_ordering
class ChemicalBase(Exposure):
    """Enum Class which indicates the classification of chemical attack."""


@dataclass(frozen=True)
class ExposureClassesBase:
    """Parent class which serves as a container for the Exposure classes.

    Exposure classes related to environmental conditions in accordance with EN 206-1
    """

    carbonation: CarbonationBase
    chloride: ChlorideBase
    chloride_seawater: ChlorideSeawaterBase
    freeze: FreezeThawBase
    chemical: ChemicalBase

    @property
    def no_risk(self) -> bool:
        """Check if all exposure classes are 'Not applicable'.

        This represents X0 class designation according to table 4.1 from NEN-EN 1992-1-1+C2:2011.

        Returns
        -------
        bool
            True if all exposure classes are 'Not applicable'
        """
        return all(exposure_class.value == "Not applicable" for exposure_class in self.__dict__.values())

    def __str__(self) -> str:
        """String representation of the ExposureClasses object.

        Returns
        -------
        str
            String representation of the ExposureClasses object
        """
        return "X0" if self.no_risk else ", ".join(enum.value for enum in self.__dict__.values() if enum.value != "Not applicable")

    def __iter__(self) -> Iterator[Exposure]:
        """Iterator for the ExposureClasses object.

        Returns
        -------
        Iterable[Exposure]
            Iterator for the ExposureClasses object
        """
        return iter(self.__dict__.values())



"""Module for validation actions inside of Blueprints."""


class LessOrEqualToZeroError(Exception):
    """Raised when a value is less than or equal to zero."""

    def __init__(self, value_name: str, value: float) -> None:
        message = f"Invalid value for '{value_name}': {value}. Values for '{value_name}' must be greater than zero."
        super().__init__(message)


class NegativeValueError(Exception):
    """Raised when a value is negative."""

    def __init__(self, value_name: str, value: float) -> None:
        message = f"Invalid value for '{value_name}': {value}. Values for '{value_name}' cannot be negative."
        super().__init__(message)


class GreaterThan90Error(Exception):
    """Raised when a value is greater than 90."""

    def __init__(self, value_name: str, value: float) -> None:
        message = f"Invalid value for '{value_name}': {value}. Values for '{value_name}' cannot be greater than 90."
        super().__init__(message)


class ListsNotSameLengthError(Exception):
    """Raised when two lists are not of the same length."""

    def __init__(self, list_name_1: str, list_name_2: str, length_1: int, length_2: int) -> None:
        message = (
            f"The lists '{list_name_1}' and '{list_name_2}' are not of the same length. "
            f"'{list_name_1}' length: {length_1}, '{list_name_2}' length: {length_2}."
        )
        super().__init__(message)


def raise_if_less_or_equal_to_zero(**kwargs: float) -> None:
    """Raise a LessOrEqualToZeroError if any of the given keyword arguments are less than or equal to zero.

    Parameters
    ----------
    **kwargs : dict[str, float]
        A dictionary of keyword arguments where keys are parameter names, and values are the values to validate.

    Raises
    ------
    LessOrEqualToZeroError
        If any value is less than or equal to zero.

    """
    for key, value in kwargs.items():
        if value <= 0:
            raise LessOrEqualToZeroError(value_name=key, value=value)


def raise_if_negative(**kwargs: float) -> None:
    """Raise a NegativeValueError if any of the given keyword arguments are negative.

    Parameters
    ----------
    **kwargs : dict[str, float]
        A dictionary of keyword arguments where keys are parameter names, and values are the values to validate.

    Raises
    ------
    NegativeValueError
        If any value is negative.

    """
    for key, value in kwargs.items():
        if value < 0:
            raise NegativeValueError(value_name=key, value=value)


def raise_if_greater_than_90(**kwargs: float) -> None:
    """Raise a GreaterThan90Error if any of the given keyword arguments are greater than 90.

    Parameters
    ----------
    **kwargs : dict[str, float]
        A dictionary of keyword arguments where keys are parameter names, and values are the values to validate.

    Raises
    ------
    GreaterThan90Error
        If any value is greater than 90.

    """
    for key, value in kwargs.items():
        if value > 90:
            raise GreaterThan90Error(value_name=key, value=value)


def raise_if_lists_differ_in_length(**kwargs: list) -> None:
    """Check if all provided lists are of the same length.

    Parameters
    ----------
    **kwargs : dict[str, list]
        A dictionary of keyword arguments where keys are list names and values are the lists to check.

    Raises
    ------
    ListsNotSameLengthError
        If any two lists are not of the same length.
    """
    # Convert the kwargs items to a list of (name, list) tuples
    lists = list(kwargs.items())

    # Compare each list with the first list
    first_list_name, first_list = lists[0]
    first_length = len(first_list)

    for list_name, lst in lists[1:]:
        if len(lst) != first_length:
            raise ListsNotSameLengthError(first_list_name, list_name, first_length, len(lst))



"""Formula 3.8 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula


class Form3Dot8TotalShrinkage(Formula):
    """Class representing formula 3.8 for the calculation of the total shrinkage."""

    label = "3.8"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        epsilon_cd: float,
        epsilon_ca: float,
    ) -> None:
        """[εcs] The total shrinkage [-].

        NEN-EN 1992-1-1+C2:2011 art.3.1.4(6) - Formula (3.8)

        Parameters
        ----------
        epsilon_cd : float
            [εcd] Drying shrinkage [-].
        epsilon_ca : float
            [εca] Autogene shrinkage [-].

        Returns
        -------
        None
        """
        super().__init__()
        self.epsilon_cd = epsilon_cd
        self.epsilon_ca = epsilon_ca

    @staticmethod
    def _evaluate(
        epsilon_cd: float,
        epsilon_ca: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        return epsilon_cd + epsilon_ca

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.8."""
        return LatexFormula(
            return_symbol=r"\epsilon_{cs}",
            result=f"{self:.3f}",
            equation=r"\epsilon_{cd} + \epsilon_{ca}",
            numeric_equation=rf"{self.epsilon_cd:.3f} + {self.epsilon_ca:.3f}",
            comparison_operator_label="=",
        )



"""Formula 6.2 from NEN-EN 1993-1-1+C2+A1:2016: Chapter 6 - Ultimate limit state."""

# pylint: disable=arguments-differ
from blueprints.codes.eurocode.nen_en_1993_1_1_c2_a1_2016 import NEN_EN_1993_1_1_C2_A1_2016
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, KNM, RATIO
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot2UtilizationRatio(Formula):
    """Class representing form 6.2 for the calculation of the utilization ratio :math:`UC`."""

    label = "6.2"
    source_document = NEN_EN_1993_1_1_C2_A1_2016

    def __init__(self, n_ed: KN, n_rd: KN, m_y_ed: KNM, m_y_rd: KNM, m_z_ed: KNM, m_z_rd: KNM) -> None:
        """
        [:math:`UC`] The calculation of the utilization ratio [:math:`-`].

        NEN-EN 1993-1-1+C2+A1:2016 art.6.2.1(7) - Formula (6.2)

        Parameters
        ----------
        n_ed : kN
            [:math:`N_{Ed}`] Contains the design axial force [:math:`kN`].
        n_rd : kN
            [:math=`N_{Rd}`] Contains the design axial resistance [:math=`kN`].
        m_y_ed : kNm
            [:math=`M_{y,Ed}`] Contains the design moment about the y-axis [:math=`kNm`].
        m_y_rd : kNm
            [:math=`M_{y,Rd}`] Contains the design moment resistance about the y-axis [:math=`kNm`].
        m_z_ed : kNm
            [:math=`M_{z,Ed}`] Contains the design moment about the z-axis [:math=`kNm`].
        m_z_rd : kNm
            [:math=`M_{z,Rd}`] Contains the design moment resistance about the z-axis [:math=`kNm`].

        Returns
        -------
        None
        """
        super().__init__()
        self.n_ed = n_ed
        self.n_rd = n_rd
        self.m_y_ed = m_y_ed
        self.m_y_rd = m_y_rd
        self.m_z_ed = m_z_ed
        self.m_z_rd = m_z_rd

    @staticmethod
    def _evaluate(
        n_ed: KN,
        n_rd: KN,
        m_y_ed: KNM,
        m_y_rd: KNM,
        m_z_ed: KNM,
        m_z_rd: KNM,
    ) -> RATIO:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(n_rd=n_rd, m_y_rd=m_y_rd, m_z_rd=m_z_rd)
        raise_if_negative(n_ed=n_ed, m_y_ed=m_y_ed, m_z_ed=m_z_ed)

        return (n_ed / n_rd) + (m_y_ed / m_y_rd) + (m_z_ed / m_z_rd)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for form 6.2."""
        numeric_equation = (
            rf"\frac{{{self.n_ed:.3f}}}{{{self.n_rd:.3f}}} + "
            rf"\frac{{{self.m_y_ed:.3f}}}{{{self.m_y_rd:.3f}}} + "
            rf"\frac{{{self.m_z_ed:.3f}}}{{{self.m_z_rd:.3f}}}"
        )
        return LatexFormula(
            return_symbol=r"UC",
            result=f"{self:.3f}",
            equation=r"\frac{N_{Ed}}{N_{Rd}} + \frac{M_{y,Ed}}{M_{y,Rd}} + \frac{M_{z,Ed}}{M_{z,Rd}}",
            numeric_equation=numeric_equation,
            comparison_operator_label="=",
        )



"""Module for the abstract base class Formula."""

from abc import ABC, abstractmethod


class Formula(float, ABC):
    """Abstract base class for formulas used in the codes."""

    def __new__(cls, *args, **kwargs) -> "Formula":
        """Method for creating a new instance of the class."""
        result = cls._evaluate(*args, **kwargs)
        instance = float.__new__(cls, result)
        instance._initialized = False  # noqa: SLF001
        return instance

    def __init__(self, *args, **kwargs) -> None:
        """Method for initializing a new instance of the class."""
        super().__init__(*args, **kwargs)
        self._initialized = True

    def __setattr__(self, name: str, value: str | float) -> None:
        """Override the __setattr__ method to prevent modifications after initialization.

        Parameters
        ----------
        name : str
            The name of the attribute.
        value : str | float
            The value to be assigned to the attribute.
        """
        if getattr(self, "_initialized", False) and name in self.__dict__:
            raise AttributeError(f"Attribute '{name}' of '{type(self).__name__}' object is read-only and cannot be modified after initialization.")
        super().__setattr__(name, value)

    @property
    @abstractmethod
    def label(self) -> str:
        """Property for the formula label.

        For example, "5.2" for formula 5.2.

        Returns
        -------
        str
            The label/number associated with the formula.
            This is an abstract method and must be implemented in all subclasses.
        """

    @property
    @abstractmethod
    def source_document(self) -> str:
        """Property for the source document.

        For example, "NEN-EN 1992-1-1+C2:2011"
        Try to use the official and complete name of the document including publishing year, if possible.

        Returns
        -------
        str
            The reference to the document where the formula originates.
            This is an abstract method and must be implemented in all subclasses.
        """

    @property
    def detailed_result(self) -> dict:
        """Property for providing the detailed result of the formula.

        Returns
        -------
        dict
            The detailed result of the formula.
            Keys are strings representing the name of the partial or intermediate result.
            Values types will depend on the specific implementation, but must be a serializable type.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _evaluate(*args, **kwargs) -> float:
        """Abstract method for the logic of the formula.

        Returns
        -------
        float
            The result of the formula.
            This is an abstract method and must be implemented in all subclasses.
        """



"""Formula 3.23 from NEN-EN 1992-1-1+C2:2011: Chapter 3 - Materials."""

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM, MPA


class Form3Dot23FlexuralTensileStrength(Formula):
    """Class representing formula 3.23 for the calculation of the mean flexural tensile strength of reinforced concrete members."""

    label = "3.23"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        h: MM,
        f_ctm: MPA,
    ) -> None:
        """[fctm,fl] Mean flexural tensile strength of reinforced concrete members  [MPa].

        NEN-EN 1992-1-1+C2:2011 art.3.1.8(1) - Formula (3.23)

        Parameters
        ----------
        h : MM
            [h] Total member depth [mm].
        f_ctm : MPA
            [fctm] Mean axial tensile strength following from table 3.1 [MPa].

        Returns
        -------
        None
        """
        super().__init__()
        self.h = h
        self.f_ctm = f_ctm

    @staticmethod
    def _evaluate(
        h: MM,
        f_ctm: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if h < 0:
            raise ValueError(f"Invalid h: {h}. h cannot be negative")
        if f_ctm < 0:
            raise ValueError(f"Invalid f_ctm: {f_ctm}. f_ctm cannot be negative")
        return max((1.6 - h / 1000) * f_ctm, f_ctm)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 3.23."""
        return LatexFormula(
            return_symbol=r"f_{ctm,fl}",
            result=f"{self:.3f}",
            equation=r"\max \left[ (1.6 - h/1000) \cdot f_{ctm} ; f_{ctm} \right]",
            numeric_equation=rf"\max \left[ (1.6 - {self.h:.3f}/1000) \cdot {self.f_ctm:.3f} ; {self.f_ctm:.3f} \right]",
            comparison_operator_label="=",
        )



"""Formula 5.9 from NEN-EN 1993-5:2008 Chapter 5 - Ultimate limit state."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1993_5_2008 import NEN_EN_1993_5_2008
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_fraction, latex_min_curly_brackets
from blueprints.type_alias import DEG, DIMENSIONLESS, KNM, MM, MM2, MM3, MPA
from blueprints.unit_conversion import NMM_TO_KNM
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot9ReducedBendingMomentResistance(Formula):
    """Class representing formula 5.9 for reduced design bending moment resistance of the cross-section."""

    label = "5.9"
    source_document = NEN_EN_1993_5_2008

    def __init__(
        self,
        beta_b: DIMENSIONLESS,
        w_pl: MM3,
        rho: DIMENSIONLESS,
        a_v: MM2,
        t_w: MM,
        alpha: DEG,
        f_y: MPA,
        gamma_m_0: DIMENSIONLESS,
        mc_rd: KNM,
    ) -> None:
        """(:math:`M_{V,Rd}`) Calculate reduced design bending moment resistance of the cross-section allowing for the shear force in [:math:`kNm`].

        This calculation is specifically for sheet pile cross-sections, particularly U-profiles and Z-profiles.

        NEN-EN 1993-5:2008(E) art.5.2.2(9) - Formula (5.9)

        Parameters
        ----------
        beta_b : DIMENSIONLESS
            (:math:`β_{b}`) Reduction factor for the bending resistance of the cross-section, which takes account of
            possible lack of shear force transmission in the interlocks [-].
            Defined in NEN-EN 1993-5:2008(E) art. 5.2.2(2) or CUR166, part 2, par. 3.3.2.
        w_pl : MM3
            (:math:`W_{pl}`) Plastic section modulus in [:math:`mm³`].
        rho : DIMENSIONLESS
            (:math:`ρ`) Reduction factor for shear resistance of the cross-section, according NEN-EN 1993-5:2008(E) art. 5.2.2(9) formula 5.10 [-].
        a_v : MM2
            (:math:`A_{V}`) Projected shear area for each web, acting in the same direction as VEd in [:math:`mm²`].
        t_w : MM
            (:math:`t_{w}`) Thickness of the web in [:math:`mm`].
        alpha : DEGREE
            (:math:`α`) the inclination of the web according to NEN-EN 1993-5:2008(E) Figure 5-1 in [degrees].
        f_y : MPA
            (:math:`f_{y}`) Yield strength in [:math:`MPa`].
        gamma_m_0 : DIMENSIONLESS
            (:math:`γ_{M0}`) Partial factor for material properties in [-].
        mc_rd : KNM
            (:math:`M_{c,Rd}`) Design moment resistance of the cross-section in [:math:`kNm`].

            The `mc_rd` parameter represents the design moment resistance of the cross-section.
            In the context of the formula for reduced bending moment resistance, it serves as an upper bound.
            The formula calculates the reduced design bending moment resistance (`m_v_rd`) and then returns the minimum of `m_v_rd` and `mc_rd`.
            This means that the result of the formula will never exceed `mc_rd`, making `mc_rd` an upper bound for this formula.

            :math:`M_{v,Rd} <= M_{c,Rd}`
        """
        super().__init__()
        self.beta_b: float = beta_b
        self.w_pl: float = w_pl
        self.rho: float = rho
        self.a_v: float = a_v
        self.t_w: float = t_w
        self.alpha: float = alpha
        self.f_y: float = f_y
        self.gamma_m_0: float = gamma_m_0
        self.mc_rd: float = mc_rd

    @staticmethod
    def _evaluate(
        beta_b: DIMENSIONLESS,
        w_pl: MM3,
        rho: DIMENSIONLESS,
        a_v: MM2,
        t_w: MM,
        alpha: DEG,
        f_y: MPA,
        gamma_m_0: DIMENSIONLESS,
        mc_rd: KNM,
    ) -> KNM:
        """Evaluates the formula for reduced bending moment resistance."""
        raise_if_less_or_equal_to_zero(beta_b=beta_b, w_pl=w_pl, rho=rho, a_v=a_v, t_w=t_w, alpha=alpha, f_y=f_y, gamma_m_0=gamma_m_0, mc_rd=mc_rd)
        m_v_rd = ((beta_b * w_pl) - ((rho * a_v**2) / (4.0 * t_w * np.sin(np.deg2rad(alpha))))) * (f_y / gamma_m_0) * NMM_TO_KNM
        return min(m_v_rd, mc_rd)

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 5.9."""
        latex_equation = latex_min_curly_brackets(
            r"\left(\beta_b \cdot W_{pl} - \frac{\rho \cdot A_v^2}{4 \cdot t_w \cdot \sin(\alpha)}\right) \cdot \frac{f_y}{\gamma_{M0}}, M_{c,Rd}"
        )
        fraction = latex_fraction(numerator=f"{self.f_y:.2f}", denominator=f"{self.gamma_m_0:.2f}")
        return LatexFormula(
            return_symbol=r"M_{V,Rd}",
            result=f"{self:.3f}",
            equation=latex_equation,
            numeric_equation=(
                latex_min_curly_brackets(
                    rf"\left({self.beta_b:.2f} \cdot {self.w_pl:.2f} - \frac{{{self.rho:.2f} \cdot {self.a_v:.2f}^2}}{{4 \cdot {self.t_w:.2f} \cdot "
                    rf"\sin({self.alpha:.2f})}}\right) \cdot "
                    rf"{fraction} \cdot 10^{{-6}}, {self.mc_rd:.2f}"
                )
            ),
            comparison_operator_label="=",
        )



"""Formula 9.2 from NEN-EN 1992-1-1+C2:2011: Chapter 9 - Detailing of members and particular rules."""

import numpy as np

from blueprints.codes.eurocode.nen_en_1992_1_1_c2_2011 import NEN_EN_1992_1_1_C2_2011
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DEG, MM
from blueprints.validations import raise_if_greater_than_90, raise_if_negative


class Form9Dot2ShiftInMomentDiagram(Formula):
    """Class representing the formula 9.2 for the calculation of the shift in the moment diagram for elements with shear reinforcement."""

    label = "9.2"
    source_document = NEN_EN_1992_1_1_C2_2011

    def __init__(
        self,
        z: MM,
        theta: DEG,
        alpha: DEG,
    ) -> None:
        """[al] Shift in the moment diagram of an element with shear reinforcement [mm].

        NEN-EN 1992-1-1+C2:2011 art.9.2.1.3(2) - Formula (9.2)

        Parameters
        ----------
        z: MM
            [z] The internal lever arm for an element with constant height, corresponding to the bending moment in the considered element. In the
            shear force calculation of reinforced concrete without axial force, the approximate value z = 0.9d may generally be used [mm].
        alpha: DEG
            [α] The angle between the shear reinforcement and the longitudinal axis of the beam (see 9.2.2(1)) [deg].
        theta: DEG
            [θ] The angle between the shear compression strut and the axis of the beam 6.2.3 [C1] [deg].
        """
        super().__init__()
        self.z = z
        self.theta = theta
        self.alpha = alpha

    @staticmethod
    def _evaluate(
        z: MM,
        theta: DEG,
        alpha: DEG,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(z=z, theta=theta, alpha=alpha)
        raise_if_greater_than_90(theta=theta, alpha=alpha)

        # Convert the angle from degrees to radians
        theta_radians = np.deg2rad(theta)
        alpha_radians = np.deg2rad(alpha)

        # Calculate the cotangent
        cot_theta = 1 / np.tan(theta_radians)
        cot_alpha = 1 / np.tan(alpha_radians)

        return z * (cot_theta - cot_alpha) / 2

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for formula 9.2."""
        return LatexFormula(
            return_symbol=r"a_l",
            result=f"{self:.2f}",
            equation=r"z \cdot \left( \cot(\theta) - \cot(\alpha) \right) / 2",
            numeric_equation=rf"{self.z:.2f} \cdot \left( \cot({self.theta:.2f}) - \cot({self.alpha:.2f}) \right) / 2",
            comparison_operator_label="=",
        )



"""Module for the representation of reinforcement configurations in reinforced concrete sections."""

from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
from shapely import LineString

from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial
from blueprints.structural_sections.concrete.rebar import Rebar
from blueprints.type_alias import DIMENSIONLESS, MM, MM2, MM2_M
from blueprints.unit_conversion import M_TO_MM


@dataclass(frozen=True)
class ReinforcementConfiguration(ABC):
    """Base class of all reinforcement configurations.

    Parameters
    ----------
    diameter : MM
        Diameter of the rebar [mm].
    material : ReinforcementSteelMaterial
        Representation of the properties of reinforcement steel suitable for use with NEN-EN 1992-1-1.
    """

    diameter: MM
    material: ReinforcementSteelMaterial

    def __post_init__(self) -> None:
        """Post initialization of the reinforcement configuration."""
        # diameter is a positive number
        if self.diameter <= 0:
            msg = f"Diameter of the rebar must be a positive number, got {self.diameter}"
            raise ValueError(msg)

    @property
    @abstractmethod
    def area(self) -> MM2:
        """Each reinforcement configuration must have a resulting area."""

    @abstractmethod
    def to_rebars(self, line: LineString) -> list[Rebar]:
        """Convert the reinforcement configuration to a list of rebars.

        Parameters
        ----------
        line : LineString
            Representing the path of the reinforcement in the section.
            Start of the line defines the first rebar of the configuration, end of the line defines the last rebar.

        Returns
        -------
        List[Rebar]
            List of Rebar objects.
        """


@dataclass(kw_only=True, frozen=True)
class ReinforcementByDistance(ReinforcementConfiguration):
    """Representation of a reinforcement configuration given by center-to-center distance.
    For example ⌀16-150, ⌀20-200, ⌀25-250, ⌀32-300, etc.

    Parameters
    ----------
    center_to_center : MM
        Maximum center-to-center distance between rebars [mm].
    """

    center_to_center: MM

    def __post_init__(self) -> None:
        """Post initialization of the reinforcement configuration."""
        super().__post_init__()
        self._validations()

    def _validations(self) -> None:
        """Validation of the reinforcement configuration."""
        # center-to-center is at least the diameter
        if self.center_to_center < self.diameter:
            msg = f"Center-to-center distance must be at least the diameter of the rebar, got {self.center_to_center}"
            raise ValueError(msg)

    @property
    def area(self) -> MM2_M:
        """Area of the reinforcement configuration per meter [mm²/m]."""
        return 0.25 * np.pi * self.diameter**2 * (M_TO_MM / self.center_to_center)

    @property
    def n_rebars_per_meter(self) -> DIMENSIONLESS:
        """Number of rebars per meter [1/m]."""
        return 1.0 * M_TO_MM / self.center_to_center

    def to_rebars(self, line: LineString) -> list[Rebar]:
        """Convert the reinforcement configuration to a list of rebars.

        Parameters
        ----------
        line : LineString
            Representing the path of the reinforcement in the section.
            Start of the line defines the first rebar of the configuration, end of the line defines the last rebar.

        Returns
        -------
        List[Rebar]
            List of Rebar objects.
        """
        # max(int(n), 1) is used to ensure that at least one rebar is placed
        rebars = []

        # define the number of rebars based on the length of the line, minimum 1
        n_rebars = line.length / self.center_to_center
        n_rebars_applied = max(int(n_rebars), 1)

        # calculate the space between the rebars
        side_buffer = (line.length - (n_rebars_applied - 1) * self.center_to_center) / 2
        distances = np.linspace(start=side_buffer, stop=line.length - side_buffer, num=n_rebars_applied)

        # define the representative diameter of the rebar
        reinforcement_area = 0.25 * np.pi * self.diameter**2 * n_rebars
        repr_diameter = np.sqrt(reinforcement_area / (0.25 * np.pi * n_rebars_applied))

        for distance in distances:
            point = line.interpolate(distance)
            rebars.append(
                Rebar(
                    diameter=repr_diameter,
                    x=point.x,
                    y=point.y,
                    material=self.material,
                )
            )
        return rebars

    def __repr__(self) -> str:
        """Representation of the reinforcement configuration."""
        return f"{self.__class__.__name__}|{self!s}|{self.area:.0f} mm²/m"

    def __str__(self) -> str:
        """String representation of the reinforcement configuration."""
        return f"⌀{self.diameter:.0f}-{self.center_to_center:.0f}"


@dataclass(kw_only=True, frozen=True)
class ReinforcementByQuantity(ReinforcementConfiguration):
    """Representation of a reinforcement configuration given by quantity of rebars.
    For example 4⌀16, 6⌀20, 8⌀25, 10⌀32, etc.

    Parameters
    ----------
    n : int
        Amount of longitudinal bars.
    """

    n: int

    def __post_init__(self) -> None:
        """Post initialization of the reinforcement configuration."""
        super().__post_init__()
        self._validations()

    def _validations(self) -> None:
        """Validation of the reinforcement configuration."""
        # check that n is an integer
        if not isinstance(self.n, int):
            msg = f"Number of rebars must be an integer, got {self.n}"
            raise TypeError(msg)

        # check that n is at least 2
        minimum_number_of_rebars = 1
        if self.n < minimum_number_of_rebars:
            msg = f"Number of rebars must be at least {minimum_number_of_rebars}, got {self.n}"
            raise ValueError(msg)

    @property
    def area(self) -> MM2:
        """Area of the reinforcement configuration [mm²]."""
        return 0.25 * np.pi * self.diameter**2 * self.n

    def to_rebars(self, line: LineString) -> list[Rebar]:
        """Convert the reinforcement configuration to a list of rebars.

        Parameters
        ----------
        line : LineString
            Representing the path of the reinforcement in the section.
            Start of the line defines the first rebar of the configuration, end of the line defines the last rebar.

        Returns
        -------
        List[Rebar]
            List of Rebar objects.
        """
        rebars = []
        for index in range(self.n):
            distance = index * line.length / (self.n - 1)
            point = line.interpolate(distance)
            rebars.append(
                Rebar(
                    diameter=self.diameter,
                    x=point.x,
                    y=point.y,
                    material=ReinforcementSteelMaterial(),
                )
            )
        return rebars

    def __repr__(self) -> str:
        """Representation of the reinforcement by quantity."""
        return f"{self.__class__.__name__}|{self!s}|{self.area:.0f} mm²"

    def __str__(self) -> str:
        """String representation of the reinforcement by quantity."""
        return f"{self.n}⌀{self.diameter:.0f}"
